#include <suifkernel/suif_env.h>
#include <suifkernel/module_subsystem.h>
#include <suifkernel/module.h>
#include <suifkernel/command_line_parsing.h>
#include <suifpasses/passes.h>
#include <stdio.h>

#include <iostream>
#include <fstream>

#include <basicnodes/basic.h>
#include <basicnodes/basic_factory.h>

#include <suifnodes/suif.h>
#include <suifnodes/suif_factory.h>

#include <esuifnodes/esuif.h>
#include <esuifnodes/esuif_factory.h>

#include <typebuilder/type_builder.h>


#include "EsterelParser.hpp"
#include "EsterelLexer.hpp"
#include "EsterelTreeParser.hpp"
#include "LineAST.hpp"
#include <antlr/AST.hpp>

extern "C" void init_utils(SuifEnv *);

extern void strl2suif_parse_file(const char *, FileSetBlock *, SuifEnv *);

static String _file_name;
static String _ast_file_name;
static const String _dash_string("-");
static const String _empty_string("");

class Strl2suifPass : public FrontendPass {
public:
  Strl2suifPass(SuifEnv *env, const LString &name)
    : FrontendPass(env, name) {}

  ~Strl2suifPass() {}

  Module *clone() const;
  void initialize();

  /// Does the work: called by FrontendPass::execute()
  FileSetBlock *build_file_set_block();
};

Module *Strl2suifPass::clone() const
{
  return (Module*)this;
}

void Strl2suifPass::initialize()
{
  FrontendPass::initialize();


  Option *ast_file_option =
    build_prefixed_string("-ast", "AST filename",
			  &_ast_file_name,
			  "name of AST file to generate, - for stdout");
  OptionSelection *opt = new OptionSelection(true);
  opt->add(ast_file_option);
  _command_line->add(opt);

  _command_line->set_description("Create SUIF for an Esterel source file.");
  OptionString *_file_name_argument =
    new OptionString("Filename", &_file_name);
  _file_name_argument->set_description("Input file name");
  _command_line->add(_file_name_argument);

}

void printAST(std::ostream &os, RefLineAST ast, int indent = 0)
{
  while (ast) {
    for (int i = 0 ; i < indent ; i++ ) os << "  ";
    os << ast->toString() << " <" << ast->getType() << '>' << endl;
    printAST(os, ast->getFirstChild(), indent+1);
    ast = ast->getNextSibling();
  }
}

FileSetBlock *Strl2suifPass::build_file_set_block()
{
  // printf("build_file_set_block called: filename %s\n", _file_name.c_str());

  SuifEnv *env = get_suif_env();

  FileSetBlock *fsb = create_file_set_block(env);
  env->set_file_set_block(fsb);
  suif_assert(fsb);

  // Create the built-in types

  TypeBuilder *tb = get_type_builder(env);
  suif_assert(tb);

  EsterelInformationBlock *eib = create_esterel_information_block(env);
  fsb->append_information_block(eib);

  // Construct the built-in types
  eib->set_integer_type(tb->get_integer_type(32, 32, true));
  eib->set_q_integer_type(tb->get_qualified_type(eib->get_integer_type()));
  eib->set_boolean_type(tb->get_boolean_type(32, 32));
  eib->set_q_boolean_type(tb->get_qualified_type(eib->get_boolean_type()));
  eib->set_string_type(tb->get_pointer_type(32, 32,
					    tb->get_integer_type(8, 8, true)));
  eib->set_float_type(tb->get_floating_point_type(32, 32));
  eib->set_double_type(tb->get_floating_point_type(64, 64));
  eib->set_void_type(tb->get_void_type());
  eib->set_q_void_type(tb->get_qualified_type(eib->get_void_type()));

  try {
    ifstream is(_file_name.c_str());
    EsterelLexer lexer(is);
    EsterelParser parser(lexer);
    parser.setASTNodeFactory( LineAST::factory );
    // std::cout << "Parsing " << _file_name << endl;
    parser.file();
    // std::cout << "Done parsing" << endl;
    RefLineAST t = parser.getAST();
    //   std::cout << t->toStringTree() << endl;

    /// Print the AST if requested

    if (_ast_file_name == _dash_string) {
      printAST(std::cout,t);
    } else if (_ast_file_name != _empty_string ) {
      ofstream of(_ast_file_name);
      printAST(of, t);
      of.close();
    }

    /// Walk the AST and build the FileBlock

    EsterelTreeParser treeParser;
    treeParser.set_suif_env(get_suif_env());
    treeParser.file(t, fsb, _file_name);
    if ( treeParser.get_num_errors() > 0 ) {
       suif_error("Compilation aborted due to errors");      
    }

  } catch (std::exception& e) {
    std::cerr << "exception: " << e.what() << std::endl;
  }

  return fsb;
}

extern "C" void init_strl2suif(SuifEnv *suif_env)
{
  init_basicnodes(suif_env);
  init_suifnodes(suif_env);
  init_esuifnodes(suif_env);
  init_utils(suif_env);
  init_typebuilder(suif_env);
  // Add more modules to initialize (?)
  ModuleSubSystem *ms = suif_env->get_module_subsystem();
  if (!ms->retrieve_module("strl2suif"))
    ms->register_module(new Strl2suifPass(suif_env, "strl2suif"));
}
