#include "statement_dismantlers.h"
#include <suifkernel/command_line_parsing.h>

/** Generic pass that invokes a walker class.
    Taken from basepases/transforms/transforms.cpp */

template <class walker> class WalkerModule : public PipelinablePass {
  String _description;
public:
  WalkerModule(SuifEnv *env, const LString &name,
		      const String &description ) 
    : PipelinablePass(env, name), _description(description) {}
    
  virtual ~WalkerModule()  {}

  void initialize() {
    PipelinablePass::initialize();
    _command_line->set_description(_description);
  }

  virtual void do_procedure_definition(ProcedureDefinition *proc_def) {
    if ( !proc_def ) return;
    walker walk(get_suif_env(), proc_def);
    proc_def->walk(walk);
  }

  Module *clone() const { return(Module*)this; }
};

extern "C" void init_transformstrl(SuifEnv *suif_env)
{
  suif_env->require_module("suifnodes");
  suif_env->require_module("esuifnodes");

  init_basicnodes(suif_env);
  init_suifnodes(suif_env);
  init_esuifnodes(suif_env);
  init_typebuilder(suif_env);

  ModuleSubSystem *ms = suif_env->get_module_subsystem();

  ms->register_module(new WalkerModule< Dismantler<PresentStatement> >(
     suif_env, "dismantle_present_statements",
     "Dismantle PresentStatements to IfStatements"));

  ms->register_module(new WalkerModule< Dismantler<EsterelIfStatement> >(
     suif_env, "dismantle_esterel_if_statements",
     "Dismantle EsterelIfStatements to IfStatements"));

  ms->register_module(new WalkerModule< Dismantler<AwaitStatement> >(
     suif_env, "dismantle_await_statements",
     "Dismantle AwaitStatements to AbortStatements"));

  ms->register_module(new WalkerModule< Dismantler<AbortStatement> >(
     suif_env, "dismantle_abort_statements",
     "Dismantle AbortStatements to ResumeStatements"));

  ms->register_module(new WalkerModule< Dismantler<DoWatchingStatement> >(
     suif_env, "dismantle_do_watching_statements",
     "Dismantle DoWatchingStatements to AbortStatements"));

  ms->register_module(new WalkerModule< Dismantler<DoUptoStatement> >(
     suif_env, "dismantle_do_upto_statements",
     "Dismantle DoUptoStatements to AbortStatements"));

  ms->register_module(new WalkerModule< Dismantler<LoopEachStatement> >(
     suif_env, "dismantle_loop_each_statements",
     "Dismantle LoopEachStatements to AbortStatements"));

  ms->register_module(new WalkerModule< Dismantler<HaltStatement> >(
     suif_env, "dismantle_halt_statements",
     "Dismantle HaltStatements to loop and pause"));

  ms->register_module(new WalkerModule< Dismantler<SustainStatement> >(
     suif_env, "dismantle_sustain_statements",
     "Dismantle SustainStatements to loop and emit"));

  ms->register_module(new WalkerModule< Dismantler<LoopStatement> >(
     suif_env, "dismantle_loop_statements",
     "Dismantle LoopStatements to a label and goto"));

  ms->register_module(new ParallelTrapExitModule(
     suif_env, "dismantle_parallel_trap_exit_statements",
     "Dismantle parallel, trap, and exit statements."));

  ms->register_module(new WalkerModule< Dismantler<LoadTrapExpression> >(
     suif_env, "dismantle_load_trap_expressions",
     "Dismantle loads of trap presence/absence into variable expression."));

  ms->register_module(new WalkerModule< Dismantler<EmitStatement> >(
     suif_env, "dismantle_emit_statements",
     "Dismantle emit statements into variable stores."));

  ms->register_module(new WalkerModule< Dismantler<ThreadStatement> >(
     suif_env, "dismantle_thread_statements",
     "Dismantle thread statements into their bodies."));

  /*
  ms->register_module(new FixProcDefs(
     suif_env, "fix_esterel_proc_defs",
     "Transform Esterel modules into reasonable ProcedureDefinitions."));
  */

  ms->register_module(new WalkerModule< Dismantler<PauseStatement> >(
     suif_env, "dismantle_pause_statements",
     "Dismantle pause statements into constant assignments and branches."));

  ms->register_module(new WalkerModule< Dismantler<ResumeStatement> >(
     suif_env, "dismantle_resume_statements",
     "Dismantle resume statements into multi-way branches."));

  ms->register_module(new WalkerModule< Dismantler<NothingStatement> >(
     suif_env, "dismantle_nothing_statements",
     "Dismantle nothing statements into empty sequences."));

  ms->register_module(new WalkerModule< Dismantler<ExitAtStatement> >(
     suif_env, "dismantle_exit_at_statements",
     "Dismantle ExitAt statements into assignments and gotos."));

  ms->register_module(new BuildCInterface(
     suif_env, "build_c_interface",
     "Build the functions for the C interface to an Esterel module."));

#if 0
  ms->register_module(new ParallelTrapStatementDismantler(
     suif_env, "dismantle_parallel_trap_statements",
     "Dismantle trap, exit, and parallel statements"));
#endif
}
